# spec format in dictionary:
# {
#   'size'                  : int,
#   'should_remove_alpha'   : bool,
#   'should_crop_to_rounded': bool,
#   'crop_height'           : Optional[int],
# }

icon_spec = {
    "ios": {
        "AppIcon.appiconset": {
            "Icon-App-20x20@1x.png": {
                "size": 20,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-20x20@2x.png": {
                "size": 40,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-20x20@3x.png": {
                "size": 60,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-29x29@1x.png": {
                "size": 29,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-29x29@2x.png": {
                "size": 58,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-29x29@3x.png": {
                "size": 87,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-40x40@1x.png": {
                "size": 40,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-40x40@2x.png": {
                "size": 80,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-40x40@3x.png": {
                "size": 120,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-60x60@2x.png": {
                "size": 120,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-60x60@3x.png": {
                "size": 180,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-76x76@1x.png": {
                "size": 76,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-76x76@2x.png": {
                "size": 152,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-App-83.5x83.5@2x.png": {
                "size": 167,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ItunesArtwork@2x.png": {
                "size": 1024,
                "should_remove_alpha": True,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
        },
        # iTunes artwork
        "iTunesArtwork@1x.png": {
            "size": 512,
            "should_remove_alpha": True,
            "should_crop_to_rounded": False,
            "crop_height": None,
        },
        "iTunesArtwork@2x.png": {
            "size": 1024,
            "should_remove_alpha": True,
            "should_crop_to_rounded": False,
            "crop_height": None,
        },
        "iTunesArtwork@3x.png": {
            "size": 1536,
            "should_remove_alpha": True,
            "should_crop_to_rounded": False,
            "crop_height": None,
        },
    },
    "imessenger": {
        "icon-messages-app-27x20@1x.png": {
            "size": 27,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 20,
        },
        "icon-messages-app-27x20@2x.png": {
            "size": 54,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 40,
        },
        "icon-messages-app-27x20@3x.png": {
            "size": 81,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 60,
        },
        "icon-messages-app-iPadAir-67x50@2x.png": {
            "size": 134,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 100,
        },
        "icon-messages-app-iPadAir-74x55@2x.png": {
            "size": 148,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 110,
        },
        "icon-messages-app-iPhone-60x45@1x.png": {
            "size": 60,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 45,
        },
        "icon-messages-app-iPhone-60x45@2x.png": {
            "size": 120,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 90,
        },
        "icon-messages-app-iPhone-60x45@3x.png": {
            "size": 180,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 135,
        },
        "icon-messages-settings-29x29@2x.png": {
            "size": 58,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 58,
        },
        "icon-messages-settings-29x29@3x.png": {
            "size": 87,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 87,
        },
        "icon-messages-transcript-32x24@1x.png": {
            "size": 32,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 24,
        },
        "icon-messages-transcript-32x24@2x.png": {
            "size": 64,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 48,
        },
        "icon-messages-transcript-32x24@3x.png": {
            "size": 96,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 72,
        },
        "icon-messages-app-store-1024x1024.png": {
            "size": 1024,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 1024,
        },
        "icon-messages-app-store-1024x768.png": {
            "size": 1024,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": 768,
        },
    },
    "watchkit": {
        "AppIcon.appiconset": {
            # WatchKit Notification Center icon
            "Icon-24@2x.png": {
                "size": 48,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-27.5@2x.png": {
                "size": 55,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            # WatchKit Long-Look notification icon
            "Icon-40@2x.png": {
                "size": 80,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-44@2x.png": {
                "size": 88,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-50@2x.png": {
                "size": 100,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            # WatchKit App icon
            "Icon-29@2x.png": {
                "size": 58,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-29@3x.png": {
                "size": 87,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            # WatchKit Short-Look icon
            "Icon-86@2x.png": {
                "size": 172,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-98@2x.png": {
                "size": 196,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "Icon-108@2x.png": {
                "size": 216,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ItunesArtwork@2x.png": {
                "size": 1024,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
        }
    },
    "android": {
        # TODO: Keep this? Android docs says it is resized by the system
        "mipmap-mdpi": {
            "ic_launcher.png": {
                "size": 48,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ic_launcher_foreground.png": {
                "size": 108,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ic_launcher_round.png": {
                "size": 48,
                "should_remove_alpha": False,
                "should_crop_to_rounded": True,
                "crop_height": None,
            },
        },
        "mipmap-hdpi": {
            "ic_launcher.png": {
                "size": 72,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ic_launcher_foreground.png": {
                "size": 162,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ic_launcher_round.png": {
                "size": 72,
                "should_remove_alpha": False,
                "should_crop_to_rounded": True,
                "crop_height": None,
            },
        },
        "mipmap-xhdpi": {
            "ic_launcher.png": {
                "size": 96,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ic_launcher_foreground.png": {
                "size": 216,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ic_launcher_round.png": {
                "size": 96,
                "should_remove_alpha": False,
                "should_crop_to_rounded": True,
                "crop_height": None,
            },
        },
        "mipmap-xxhdpi": {
            "ic_launcher.png": {
                "size": 144,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ic_launcher_foreground.png": {
                "size": 324,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ic_launcher_round.png": {
                "size": 144,
                "should_remove_alpha": False,
                "should_crop_to_rounded": True,
                "crop_height": None,
            },
        },
        "mipmap-xxxhdpi": {
            "ic_launcher.png": {
                "size": 192,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ic_launcher_foreground.png": {
                "size": 432,
                "should_remove_alpha": False,
                "should_crop_to_rounded": False,
                "crop_height": None,
            },
            "ic_launcher_round.png": {
                "size": 192,
                "should_remove_alpha": False,
                "should_crop_to_rounded": True,
                "crop_height": None,
            },
        },
        "playstore-icon.png": {
            "size": 512,
            "should_remove_alpha": False,
            "should_crop_to_rounded": False,
            "crop_height": None,
        },
    },
}



import io
import logging
import zipfile
import os

from PIL import Image, ImageDraw
from js import Uint8Array, File, document, window

from .spec import icon_spec

log = logging.getLogger(__name__)
usage_log = logging.getLogger("mki_usage")


class Processor:
    def __init__(self, png_img):
        self.png_img = png_img
        self.path_stream_tuples = []

    @property
    def method_map(self):
        return {
            "should_crop_to_rounded": self.crop_to_rounded,
            "should_remove_alpha": self.remove_alpha,
            "crop_height": self.crop_height,
            "size": self.resize,
        }

    def remove_alpha(self, img, spec):
        """
        - Convert this to RGBA if possible
        - Black background canvas (r,g,b,a)
        - Paste the image onto the canvas, using it's alpha channel as mask
        """
        img.convert("RGBA")
        canvas = Image.new("RGBA", img.size, (0, 0, 0, 255))
        canvas.paste(img, mask=img)
        canvas.thumbnail([spec["size"], spec["size"]], Image.ANTIALIAS)
        img = canvas.convert("RGB")
        return img

    def crop_height(self, img, spec):
        """
        - Crop icon height for rectangular icon
        """
        crop_height_spec = spec["crop_height"]
        width, height = img.size
        top = (height - crop_height_spec) // 2
        bottom = (height + crop_height_spec) // 2
        img = img.crop((0, top, width, bottom))
        return img

    def crop_to_rounded(self, img, spec):
        """
        - Crop to rounded icon
        """
        circle = Image.new("L", img.size, 0)
        draw = ImageDraw.Draw(circle)
        draw.ellipse((0, 0) + img.size, fill=255)
        alpha = Image.new("L", img.size, 255)
        alpha.paste(circle)
        img.putalpha(alpha)
        return img

    def resize(self, img, spec):
        """
        - Resize icon to given size
        """
        img = img.resize((spec["size"], spec["size"]), Image.ANTIALIAS)
        return img

    def process(self, file_path, file_name, spec):
        img = self.png_img.copy()
        try:
            # The order in this list matters
            for pipeline in [
                "should_crop_to_rounded",
                "should_remove_alpha",
                "size",
                "crop_height",
            ]:
                if pipeline in spec and bool(spec[pipeline]) is True:
                    img = self.method_map[pipeline](img, spec)

        except ValueError:
            pass
        finally:
            if not os.path.exists(file_path):
                os.makedirs(file_path)
            new_stream = io.BytesIO()
            img.save(new_stream, format="PNG")
            self.path_stream_tuples.append((file_path + "/" + file_name, new_stream))


async def icon(origin_image_file):
    # Generate the icon.
    png_img = await convert_to_png(origin_image_file)

    path_stream_tuples = walk_spec("/icons", png_img)

    zip_file = zip_images(path_stream_tuples)
    download_zip(zip_file)

    usage_log.info("Icon generated: {}".format(png_img))


def walk_spec(file_path, png_img, spec=icon_spec):
    processor = Processor(png_img)
    path_stream_tuples = []

    for k, v in spec.items():
        if isinstance(v, dict) and "size" in v:
            processor.process(file_path, k, v)
        elif isinstance(v, dict):
            new_path = file_path + "/" + k
            new_path_stream_tuples = walk_spec(new_path, png_img, v)
            path_stream_tuples = path_stream_tuples + new_path_stream_tuples

    path_stream_tuples = path_stream_tuples + processor.path_stream_tuples
    return path_stream_tuples


def zip_images(path_stream_tuples):
    zip_file_bytes_io = io.BytesIO()
    with zipfile.ZipFile(zip_file_bytes_io, "w") as zip_file:
        for image_name, bytes_stream in path_stream_tuples:
            zip_file.writestr(image_name, bytes_stream.getvalue())

    # Create a ZIP file from the zip file data
    zip_file = File.new(
        [Uint8Array.new(zip_file_bytes_io.getvalue())], {"type": "application/zip"}
    )

    return zip_file


def download_zip(zip_file):
    # Create a URL for the ZIP object
    zip_url = window.URL.createObjectURL(zip_file)

    # Create a download link
    download_link = document.createElement("a")
    download_link.href = zip_url
    download_link.download = "icons.zip"
    download_link.click()


async def convert_to_png(origin_image):
    # Get the data from the files arrayBuffer as an array of unsigned bytes
    array_buf = Uint8Array.new(await origin_image.arrayBuffer())

    # BytesIO wants a bytes-like object, so convert to bytearray first
    bytes_list = bytearray(array_buf)
    origin_bytes = io.BytesIO(bytes_list)

    # Create PIL image from np array
    my_image = Image.open(origin_bytes)

    my_stream = io.BytesIO()
    my_image.save(my_stream, format="PNG")

    return my_image
